@@ -4,6 +4,7 @@ Preliminary documentation at https://github.com/ipython/ipython/wiki/IPEP-16%3A-
 """
 # Copyright (c) Jupyter Development Team.
 # Distributed under the terms of the Modified BSD License.
+import asyncio
 import json
 from textwrap import dedent
 from traceback import format_tb
@@ -14,23 +15,22 @@ try:
     from jupyter_client.jsonutil import json_default
 except ImportError:
     from jupyter_client.jsonutil import date_default as json_default
-from tornado import gen
-from tornado import web
-from tornado.concurrent import Future
+
+from concurrent.futures import Future
+
+from tornado import gen, web
 from tornado.ioloop import IOLoop
 
-from ...base.handlers import APIHandler
-from ...base.zmqhandlers import AuthenticatedZMQStreamHandler
-from ...base.zmqhandlers import (
-    deserialize_binary_message,
-    serialize_msg_to_ws_v1,
-    deserialize_msg_from_ws_v1,
-)
-from jupyter_server.utils import ensure_async
-from jupyter_server.utils import url_escape
-from jupyter_server.utils import url_path_join
 from jupyter_server.auth import authorized
+from jupyter_server.utils import ensure_async, url_escape, url_path_join
 
+from ...base.handlers import APIHandler
+from ...base.zmqhandlers import (
+    AuthenticatedZMQStreamHandler,
+    deserialize_binary_message,
+    deserialize_msg_from_ws_v1,
+    serialize_msg_to_ws_v1,
+)
 
 AUTH_RESOURCE = "kernels"
 
@@ -57,7 +57,9 @@ class MainKernelHandler(KernelsAPIHandler):
         else:
             model.setdefault("name", km.default_kernel_name)
 
-        kernel_id = await km.start_kernel(kernel_name=model["name"], path=model.get("path"))
+        kernel_id = await ensure_async(
+            km.start_kernel(kernel_name=model["name"], path=model.get("path"))
+        )
         model = await ensure_async(km.kernel_model(kernel_id))
         location = url_path_join(self.base_url, "api", "kernels", url_escape(kernel_id))
         self.set_header("Location", location)
@@ -116,7 +118,10 @@ class ZMQChannelsHandler(AuthenticatedZMQStreamHandler):
     # class-level registry of open sessions
     # allows checking for conflict on session-id,
     # which is used as a zmq identity and must be unique.
-    _open_sessions = {}
+    _open_sessions: dict = {}
+
+    _kernel_info_future: Future
+    _close_future: Future
 
     @property
     def kernel_info_timeout(self):
@@ -148,7 +153,7 @@ class ZMQChannelsHandler(AuthenticatedZMQStreamHandler):
         return protocol
 
     def __repr__(self):
-        return "%s(%s)" % (
+        return "{}({})".format(
             self.__class__.__name__,
             getattr(self, "kernel_id", "uninitialized"),
         )
@@ -177,9 +182,9 @@ class ZMQChannelsHandler(AuthenticatedZMQStreamHandler):
         # before connections are opened,
         # plus it is *very* unlikely that a busy kernel will not finish
         # establishing its zmq subscriptions before processing the next request.
-        if getattr(kernel, "execution_state") == "busy":
+        if getattr(kernel, "execution_state", None) == "busy":
             self.log.debug("Nudge: not nudging busy kernel %s", self.kernel_id)
-            f = Future()
+            f: Future = Future()
             f.set_result(None)
             return f
         # Use a transient shell channel to prevent leaking
@@ -191,8 +196,8 @@ class ZMQChannelsHandler(AuthenticatedZMQStreamHandler):
         # The IOPub used by the client, whose subscriptions we are verifying.
         iopub_channel = self.channels["iopub"]
 
-        info_future = Future()
-        iopub_future = Future()
+        info_future: Future = Future()
+        iopub_future: Future = Future()
         both_done = gen.multi([info_future, iopub_future])
 
         def finish(_=None):
@@ -205,7 +210,7 @@ class ZMQChannelsHandler(AuthenticatedZMQStreamHandler):
 
         def cleanup(_=None):
             """Common cleanup"""
-            loop.remove_timeout(nudge_handle)
+            loop.remove_timeout(nudge_handle)  # type:ignore[has-type]
             iopub_channel.stop_on_recv()
             if not shell_channel.closed():
                 shell_channel.close()
@@ -270,10 +275,10 @@ class ZMQChannelsHandler(AuthenticatedZMQStreamHandler):
 
             if not both_done.done():
                 log = self.log.warning if count % 10 == 0 else self.log.debug
-                log("Nudge: attempt %s on kernel %s" % (count, self.kernel_id))
+                log(f"Nudge: attempt {count} on kernel {self.kernel_id}")
                 self.session.send(shell_channel, "kernel_info_request")
                 self.session.send(control_channel, "kernel_info_request")
-                nonlocal nudge_handle
+                nonlocal nudge_handle  # type:ignore[misc]
                 nudge_handle = loop.call_later(0.5, nudge, count)
 
         nudge_handle = loop.call_later(0, nudge, count=0)
@@ -295,8 +300,9 @@ class ZMQChannelsHandler(AuthenticatedZMQStreamHandler):
             self.log.debug("Requesting kernel info from %s", self.kernel_id)
             # Create a kernel_info channel to query the kernel protocol version.
             # This channel will be closed after the kernel_info reply is received.
-            if self.kernel_info_channel is None:
+            if self.kernel_info_channel is None:  # type:ignore[has-type]
                 self.kernel_info_channel = km.connect_shell(self.kernel_id)
+            assert self.kernel_info_channel is not None
             self.kernel_info_channel.on_recv(self._handle_kernel_info_reply)
             self.session.send(self.kernel_info_channel, "kernel_info_request")
             # store the future on the kernel, so only one request is sent
@@ -315,7 +321,7 @@ class ZMQChannelsHandler(AuthenticatedZMQStreamHandler):
         idents, msg = self.session.feed_identities(msg)
         try:
             msg = self.session.deserialize(msg)
-        except:
+        except BaseException:
             self.log.error("Bad kernel_info reply", exc_info=True)
             self._kernel_info_future.set_result({})
             return
@@ -352,7 +358,7 @@ class ZMQChannelsHandler(AuthenticatedZMQStreamHandler):
             self._kernel_info_future.set_result(info)
 
     def initialize(self):
-        super(ZMQChannelsHandler, self).initialize()
+        super().initialize()
         self.zmq_stream = None
         self.channels = {}
         self.kernel_id = None
@@ -373,7 +379,7 @@ class ZMQChannelsHandler(AuthenticatedZMQStreamHandler):
 
     async def pre_get(self):
         # authenticate first
-        super(ZMQChannelsHandler, self).pre_get()
+        super().pre_get()
         # check session collision:
         await self._register_session()
         # then request kernel info, waiting up to a certain time before giving up.
@@ -402,11 +408,11 @@ class ZMQChannelsHandler(AuthenticatedZMQStreamHandler):
         loop = IOLoop.current()
         loop.add_timeout(loop.time() + self.kernel_info_timeout, give_up)
         # actually wait for it
-        await future
+        await asyncio.wrap_future(future)
 
     async def get(self, kernel_id):
         self.kernel_id = kernel_id
-        await super(ZMQChannelsHandler, self).get(kernel_id=kernel_id)
+        await super().get(kernel_id=kernel_id)
 
     async def _register_session(self):
         """Ensure we aren't creating a duplicate session.
@@ -415,7 +421,7 @@ class ZMQChannelsHandler(AuthenticatedZMQStreamHandler):
         This is likely due to a client reconnecting from a lost network connection,
         where the socket on our side has not been cleaned up yet.
         """
-        self.session_key = "%s:%s" % (self.kernel_id, self.session.session)
+        self.session_key = f"{self.kernel_id}:{self.session.session}"
         stale_handler = self._open_sessions.get(self.session_key)
         if stale_handler:
             self.log.warning("Replacing stale connection: %s", self.session_key)
@@ -426,7 +432,7 @@ class ZMQChannelsHandler(AuthenticatedZMQStreamHandler):
             self._open_sessions[self.session_key] = self
 
     def open(self, kernel_id):
-        super(ZMQChannelsHandler, self).open()
+        super().open()
         km = self.kernel_manager
         km.notify_connect(kernel_id)
 
@@ -468,7 +474,7 @@ class ZMQChannelsHandler(AuthenticatedZMQStreamHandler):
                     pass
                 # WebSockets don't respond to traditional error codes so we
                 # close the connection.
-                for channel, stream in self.channels.items():
+                for _, stream in self.channels.items():
                     if not stream.closed():
                         stream.close()
                 self.close()
@@ -478,7 +484,7 @@ class ZMQChannelsHandler(AuthenticatedZMQStreamHandler):
         km.add_restart_callback(self.kernel_id, self.on_restart_failed, "dead")
 
         def subscribe(value):
-            for channel, stream in self.channels.items():
+            for _, stream in self.channels.items():
                 stream.on_recv_stream(self._on_zmq_reply)
 
         connected.add_done_callback(subscribe)
@@ -514,6 +520,7 @@ class ZMQChannelsHandler(AuthenticatedZMQStreamHandler):
         ignore_msg = False
         if am:
             msg["header"] = self.get_part("header", msg["header"], msg_list)
+            assert msg["header"] is not None
             if msg["header"]["msg_type"] not in am:
                 self.log.warning(
                     'Received message of type "%s", which is not allowed. Ignoring.'
@@ -554,9 +561,9 @@ class ZMQChannelsHandler(AuthenticatedZMQStreamHandler):
             return
 
         if self.subprotocol == "v1.kernel.websocket.jupyter.org":
-            super(ZMQChannelsHandler, self)._on_zmq_reply(stream, parts)
+            super()._on_zmq_reply(stream, parts)
         else:
-            super(ZMQChannelsHandler, self)._on_zmq_reply(stream, msg)
+            super()._on_zmq_reply(stream, msg)
 
     def write_stderr(self, error_message, parent_header):
         self.log.warning(error_message)
@@ -607,7 +614,7 @@ class ZMQChannelsHandler(AuthenticatedZMQStreamHandler):
             # Increment the bytes and message count
             self._iopub_window_msg_count += 1
             if msg_type == "stream":
-                byte_count = sum([len(x) for x in msg_list])
+                byte_count = sum(len(x) for x in msg_list)
             else:
                 byte_count = 0
             self._iopub_window_byte_count += byte_count
@@ -696,7 +703,7 @@ class ZMQChannelsHandler(AuthenticatedZMQStreamHandler):
             return False
 
     def close(self):
-        super(ZMQChannelsHandler, self).close()
+        super().close()
         return self._close_future
 
     def on_close(self):
@@ -727,7 +734,7 @@ class ZMQChannelsHandler(AuthenticatedZMQStreamHandler):
         # This method can be called twice, once by self.kernel_died and once
         # from the WebSocket close event. If the WebSocket connection is
         # closed before the ZMQ streams are setup, they could be None.
-        for channel, stream in self.channels.items():
+        for _, stream in self.channels.items():
             if stream is not None and not stream.closed():
                 stream.on_recv(None)
                 stream.close()
@@ -785,7 +792,7 @@ default_handlers = [
     (r"/api/kernels", MainKernelHandler),
     (r"/api/kernels/%s" % _kernel_id_regex, KernelHandler),
     (
-        r"/api/kernels/%s/%s" % (_kernel_id_regex, _kernel_action_regex),
+        rf"/api/kernels/{_kernel_id_regex}/{_kernel_action_regex}",
         KernelActionHandler,
     ),
     (r"/api/kernels/%s/channels" % _kernel_id_regex, ZMQChannelsHandler),
